#include "i2c.h"

flash char TW_Str1[] = "Start";
flash char TW_Str2[] = "Slave Adr";
flash char TW_Str3[] = "Rep Start 1";
flash char TW_Str4[] = "Rep Slave Adr 1";
flash char TW_Str5[] = "Rep Start 2";
flash char TW_Str6[] = "send Adr High";
flash char TW_Str7[] = "send Adr Low";
flash char TW_Str8[] = "Rep Start 3";
flash char TW_Str9[] = "Rep Slave Adr 2";
flash char TW_Str10[] = "get Data";
flash char TW_Str11[] = "put Data";
flash char TW_Str12[] = "Akt unbk.";


void cli(void)
{
/* disable int */
#asm("cli")
}

void sei(void)
{
/* enable int */
#asm("sei")
}

void warte_TWINT()
{
unsigned int twi;
unsigned char twr;

for(twi=0;twi<30001;twi++) {
    twr=(TWCR & (1<<TWINT));
    if(twr == 0x80) {
        twi=30000;
        }
    }
}
unsigned char TWI_sendStart()
{
unsigned char st;

TW_Status=0;
if(TWI_DEB == 1) {
    sprintf(str, "TWI sende Start akt TWSR %02x dl=%02x ",TWSR,dl++);
    pstring(str);
    }

TWCR=(1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
warte_TWINT();
st=TWSR;
if(TWI_DEB == 1) {
    printf("TWI_Start gesendet TWSR %02x ",st);
    }
if(((st & 0xF8) != TW_START) && ((st & 0xF8) != TW_REPSTA))
    TWI_Error(7, TWI_ADR, TWS_Rep_Start);
return(st);
}

void TWI_sendStop()
{
if(TWI_DEB == 1) {
    printf("TWI_Stop sende");
    }
TWCR=(1<<TWINT)|(1<<TWSTO)|(1<<TWEN);
TW_Status = 0;
if(TWI_DEB == 1) {
    printf("TWI_Stop gesendet");
    }
}

unsigned char TWI_SlaveAdrRW(unsigned char mainadr, unsigned char subadr, unsigned char RW)
{
unsigned char adr1,st;

if(TW_Status != 0)
    return(TWSR);
adr1=0;
adr1=(subadr<<1)|RW;
adr1=mainadr|adr1;

TWI_ADR=adr1;        // Adresse PCF8574P = 0x40 PCF8574AP = 0x70
TWDR=TWI_ADR;

TWCR=(1<<TWINT)|(1<<TWEN);
warte_TWINT();
st=TWSR;
if(TWI_DEB == 1) {
    printf("TWI_SlaveAdr write %02x gesendet TWSR %02x",TWI_ADR,st);
    }
if(((st & 0xF8) != 0x18) && ((st & 0xF8) != 0x40))
    TWI_Error(2, TWI_ADR, TWS_Slave_Adr);
return(st);
}


/* Sollte ein Baustein kein ACK erwarten muss TWEA auf 0 sein
   der PCF8583 erwartet nach dem letzten Byte kein ACK. Sendet man trotzdem
   ein ACK ist hngen im Schacht
*/
unsigned char TWI_readByte(unsigned char TWEA01)
{
unsigned char st,c;

if(TW_Status != 0)
    return(TWSR);
if(TWEA01 == 1)
    st=(1<<TWINT) | (1<<TWEA) | (1<<TWEN);
else
    st=(1<<TWINT) | (1<<TWEN);
if(TWI_DEB == 1) {
    printf("TWI_read sende read TWEA %02x",st);
    }
TWCR=st;

warte_TWINT();
st=TWSR;
VarTWCR=TWCR;
VarTWSR=st;
c=TWDR;
VarTWDR=c;
if(TWEA01 == 1)
    if((st & 0xF8) != 0x50) {
        TW_Status=1;
        c=st;
        }
if(TWEA01 == 0)
    if((st & 0xF8) != 0x58) {
        TW_Status=1;
        c=st;
        }
if(TWI_DEB == 1) {
    printf("TWI_read read gesendet TWSR %02x char hex %02x",st,TWDR);
    }
if(TW_Status != 0) {
    if((st & 0xF8) != 0x50) {
        TWI_Error(9, TWI_ADR, TWS_get_Data);
        }
    }
return(c);
}

unsigned char TWI_writeByte(unsigned char c)
{
unsigned char st;

if(TW_Status != 0)
    return(TWSR);
if(TWI_DEB == 1) {
    printf("TWI_write sende byte %02x",c);
    }

TWDR=c;
TWCR=(1<<TWINT)|(1<<TWEN);
warte_TWINT();
st=TWSR;
if(TWI_DEB == 1) {
    printf("TWI_write byte gesendet TWSR %02x char %02x",st,c);
    }
if((st & 0xF8) != 0x28)
    TWI_Error(18, TWI_ADR, TWS_put_Data);
return(st);
}

/*
TWI Bit Rate Prescaler
TWPS1 	TWPS0 	Prescaler Value
0 	0 	1
0 	1 	4
1 	0 	16
1 	1 	64
*/
void TWI_init(unsigned int freq)
{
// SCL Freq = F_CPU / (16+(2*TWBR)*4^TWPS)
/* minimale Frequenz ist TWBR=255. Dies ergibt bei 16MHz 30418,25 KHz. TWPS1/0 wird nicht verwendet.
   maximale Frequenz ist TWBR=0. Dies ergibt bei 16MHz 1 MHz

   Aufruf TWI_init(unsigned int freq) freq in KHz

   TWBR wird dann errechnet.

*/
// TWBR = TW Address register
// TWBR = TW Bitrate register
// TWCR TW Control register
unsigned int i=0,j=942,f1;
unsigned long fcpu=16000;

while(1)
    {
    f1=fcpu/(16+(2*i));
    if(f1 == freq)
        break;
    else
        if(j<freq)
            break;
        else
           j=f1;
    i++;
    }
printf("\n\rInitialisiere TWI mit F=%03d TWBR=%03d",f1,i);
//TWI Bitrate Register, fr die Frequenz des TWI wichtig
TWBR=i;
//hier wird die Addresse des C festgelegt
TWAR=0xFE|(1<<TWGCE);
TW_Status=0;
//TWI Status Register, die 2 niederwertigsten sind Prescaler zur Taktberechnung
TWSR &= 0b11111100;

// globale Interrupts aktivieren
sei();

//TWI Control Register, hier wird der TWI aktiviert
TWCR=((1<<TWEA) | (1<<TWEN) | (1<<TWIE));

}

//erstes oder weiteres byte lesen
unsigned char TWI_readACK(unsigned char mainadr, unsigned char subadr)
{
unsigned char c1;
unsigned char st;
unsigned char adr1;

adr1=0;
adr1=(subadr<<1)|1;
adr1=mainadr|adr1;
TWI_ADR=adr1;        // Adresse PCF8574P = 0x40 PCF8574AP = 0x70
TW_Status=0;
dl=0;
st=TWI_sendStart();
st=TWI_SlaveAdrRW(mainadr,subadr,0);
st=TWI_sendStart();
st=TWI_SlaveAdrRW(mainadr,subadr,1);
c1=TWI_readByte(1);
i2c=c1;

TWI_sendStop();
return(c1);
}

//nur ein byte bzw. letztes byte lesen
unsigned char TWI_readNACK(unsigned char mainadr, unsigned char subadr)
{
unsigned char c1;
unsigned char st;
unsigned char adr1;

adr1=0;
adr1=(subadr<<1)|1;
adr1=mainadr|adr1;
TWI_ADR=adr1;        // Adresse PCF8574P = 0x40 PCF8574AP = 0x70
TW_Status=0;
dl=0;
st=TWI_sendStart();
st=TWI_SlaveAdrRW(mainadr,subadr,0);
st=TWI_sendStart();
st=TWI_SlaveAdrRW(mainadr,subadr,1);
c1=TWI_readByte(0);
i2c=c1;

TWI_sendStop();
return(c1);
}

unsigned char TWI_write(unsigned char mainadr, unsigned char subadr, unsigned char c)
{
unsigned char st;
unsigned char adr1;

adr1=0;
adr1=(subadr<<1)|0;
adr1=mainadr|adr1;
TWI_ADR=adr1;        // Adresse PCF8574P = 0x40 PCF8574AP = 0x70

TW_Status=0;
dl=0;
st=TWI_sendStart();
st=TWI_SlaveAdrRW(mainadr,subadr,0);
st=TWI_sendStart();
st=TWI_SlaveAdrRW(mainadr,subadr,0);
st=TWI_writeByte(c);
i2c=c;

TWI_sendStop();
return(0);
}

void TWI_Error(unsigned int nr, unsigned int TWI_ADR, unsigned int akt)
{
unsigned char ch;

if(TWI_DEB != 1)  {
    TW_Status=1;
    return;
    }

ch=akt;
if((ch >11) || (ch < 1))
    ch=12;
switch(ch) {
    case 1:
        strcpyf(str, TW_Str1); break;
    case 2:
        strcpyf(str, TW_Str2); break;
    case 3:
        strcpyf(str, TW_Str3); break;
    case 4:
        strcpyf(str, TW_Str4); break;
    case 5:
        strcpyf(str, TW_Str5); break;
    case 6:
        strcpyf(str, TW_Str6); break;
    case 7:
        strcpyf(str, TW_Str7); break;
    case 8:
        strcpyf(str, TW_Str8); break;
    case 9:
        strcpyf(str, TW_Str9); break;
    case 10:
        strcpyf(str, TW_Str10); break;
    case 11:
        strcpyf(str, TW_Str11); break;
    case 12:
        strcpyf(str, TW_Str12); break;
    }
printf("\n\r%02u Akt %02u ADR %02x TWSR %02x TWCR %02x TWBR %02x   ",nr,akt,TWI_ADR,TWSR,TWCR,TWBR);

delay_ms(500);
delay_ms(500);

TW_Status=1;
}

